import logging
from typing import TYPE_CHECKING

from qldebugger.aws import get_account_id, get_client
from qldebugger.config import get_config

from .lambda_ import run_lambda
from .message import delete_messages, receive_message

if TYPE_CHECKING:
    from aws_lambda_typing.events import SQSEvent
    from mypy_boto3_sqs.type_defs import ReceiveMessageResultTypeDef

logger = logging.getLogger(__name__)


def receive_messages_and_run_lambda(*, event_source_mapping_name: str) -> None:
    sqs = get_client('sqs')
    event_source_mapping = get_config().event_source_mapping[event_source_mapping_name]

    queue_arn = f'arn:{sqs.meta.partition}:sqs:{sqs.meta.region_name}:{get_account_id()}:{event_source_mapping.queue}'

    logger.debug('Execute %r event source mapping...', event_source_mapping_name)
    messages = receive_message(
        queue_name=event_source_mapping.queue,
        batch_size=event_source_mapping.batch_size,
        maximum_batching_window=event_source_mapping.maximum_batching_window,
    )
    event = convert_sqs_messages_to_event(
        aws_region=sqs.meta.region_name,
        event_source=f'{sqs.meta.partition}:sqs',
        event_source_arn=queue_arn,
        messages=messages,
    )
    run_lambda(lambda_name=event_source_mapping.function_name, event=event)
    delete_messages(queue_name=event_source_mapping.queue, messages=messages)


def convert_sqs_messages_to_event(
    *,
    aws_region: str,
    event_source: str,
    event_source_arn: str,
    messages: 'ReceiveMessageResultTypeDef',
) -> 'SQSEvent':
    return {
        'Records': [
            {
                'messageId': message['MessageId'],
                'receiptHandle': message['ReceiptHandle'],
                'body': message['Body'],
                'attributes': {},
                'messageAttributes': {},
                'md5OfBody': message['MD5OfBody'],
                'eventSource': event_source,
                'eventSourceARN': event_source_arn,
                'awsRegion': aws_region,
            }
            for message in messages['Messages']
        ],
    }



import logging
from typing import TYPE_CHECKING, Mapping

from qldebugger.aws import get_account_id, get_client

if TYPE_CHECKING:
    from mypy_boto3_sns.type_defs import MessageAttributeValueTypeDef
    from mypy_boto3_sqs.type_defs import ReceiveMessageResultTypeDef


logger = logging.getLogger(__name__)


def publish_message(
    *,
    topic_name: str,
    message: str,
    attributes: Mapping[str, 'MessageAttributeValueTypeDef'] = {},
) -> None:
    sns = get_client('sns')
    arn = f'arn:aws:sns:{sns.meta.region_name}:{get_account_id()}:{topic_name}'
    logger.info('Sending message to %r topic...', topic_name)
    sns.publish(
        TopicArn=arn,
        Message=message,
        MessageAttributes=attributes,
    )


def send_message(*, queue_name: str, message: str) -> None:
    sqs = get_client('sqs')
    queue_url = sqs.get_queue_url(QueueName=queue_name)['QueueUrl']
    logger.info('Sending message to %r queue...', queue_name)
    sqs.send_message(QueueUrl=queue_url, MessageBody=message)


def receive_message(
    *,
    queue_name: str,
    batch_size: int,
    maximum_batching_window: int,
) -> 'ReceiveMessageResultTypeDef':
    sqs = get_client('sqs')
    queue_url = sqs.get_queue_url(QueueName=queue_name)['QueueUrl']
    logger.debug('Receiving messages from %r queue...', queue_name)
    messages = sqs.receive_message(
        QueueUrl=queue_url,
        MaxNumberOfMessages=batch_size,
        WaitTimeSeconds=maximum_batching_window,
    )
    if 'Messages' not in messages:
        logger.warning('No messages received from %r queue', queue_name)
        raise RuntimeWarning('No messages received')
    logger.info('Receved %d messages from %r queue', len(messages['Messages']), queue_name)
    return messages


def delete_messages(*, queue_name: str, messages: 'ReceiveMessageResultTypeDef') -> None:
    sqs = get_client('sqs')
    queue_url = sqs.get_queue_url(QueueName=queue_name)['QueueUrl']
    logger.debug('Deleting messages of %r queue...', queue_name)
    sqs.delete_message_batch(
        QueueUrl=queue_url,
        Entries=[
            {
                'Id': message['MessageId'],
                'ReceiptHandle': message['ReceiptHandle'],
            }
            for message in messages['Messages']
        ],
    )
    logger.info('Deleted %d messages from %r queue', len(messages['Messages']), queue_name)


def purge_messages(*, queue_name: str) -> None:
    sqs = get_client('sqs')
    queue_url = sqs.get_queue_url(QueueName=queue_name)['QueueUrl']
    logger.info('Purging messages of %r queue...', queue_name)
    sqs.purge_queue(QueueUrl=queue_url)



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aws_lambda_typing.events import SQSEvent


def print_messages(event: 'SQSEvent', context: None) -> None:
    for message in event['Records']:
        print(message['body'])
    print(f'Total: {len(event["Records"])} messages')


class LambdaCustomError(Exception):
    def __str__(self) -> str:
        return 'Lambda execution fail'


def exec_fail(event: 'SQSEvent', context: None) -> None:
    raise LambdaCustomError



from abc import ABC, abstractmethod
from typing import Any, BinaryIO, Dict, List, NamedTuple, Optional, Tuple, Union

import tomli
from pydantic import BaseModel, Field, PositiveInt, field_validator


class ConfigAWS(BaseModel):
    profile: Optional[str] = None
    access_key_id: Optional[str] = None
    secret_access_key: Optional[str] = None
    session_token: Optional[str] = None
    region: Optional[str] = None
    endpoint_url: Optional[str] = None


class ConfigSecret(BaseModel, ABC):
    @abstractmethod
    def get_value(self) -> Union[str, bytes]:
        raise NotImplementedError


class ConfigSecretString(ConfigSecret):
    string: str

    def get_value(self) -> str:
        return self.string


class ConfigSecretBinary(ConfigSecret):
    binary: bytes

    def get_value(self) -> bytes:
        return self.binary


class ConfigTopicSubscriber(BaseModel):
    queue: str
    raw_message_delivery: bool = False
    filter_policy: Optional[str] = None


class ConfigTopic(BaseModel):
    subscribers: List[ConfigTopicSubscriber] = Field(default_factory=list)


class ConfigQueueRedrivePolicy(BaseModel):
    dead_letter_queue: str
    max_receive_count: PositiveInt


class ConfigQueue(BaseModel):
    redrive_policy: Optional[ConfigQueueRedrivePolicy] = None


class NameHandlerTuple(NamedTuple):
    module: str
    function: str


class ConfigLambda(BaseModel):
    handler: NameHandlerTuple
    environment: Dict[str, str] = Field(default_factory=dict)

    @field_validator('handler', mode='before')
    @classmethod
    def _split_handler(cls, v: Any) -> Tuple[str, str]:
        if not isinstance(v, str):
            raise ValueError('should be a str')
        if '.' not in v:
            raise ValueError('should have a module and function names')
        module, function = v.rsplit('.', maxsplit=1)
        return module, function


class ConfigEventSourceMapping(BaseModel):
    queue: str
    batch_size: int = 10
    maximum_batching_window: int = 0
    function_name: str


class Config(BaseModel):
    aws: ConfigAWS = Field(default_factory=ConfigAWS)
    secrets: Dict[str, Union[ConfigSecretString, ConfigSecretBinary]] = Field(default_factory=dict)
    topics: Dict[str, ConfigTopic] = Field(default_factory=dict)
    queues: Dict[str, ConfigQueue]
    lambdas: Dict[str, ConfigLambda]
    event_source_mapping: Dict[str, ConfigEventSourceMapping]

    @classmethod
    def from_toml(cls, fp: BinaryIO, /) -> 'Config':
        return cls.model_validate(tomli.load(fp))



import json
import logging
from typing import TYPE_CHECKING, Dict

from botocore.exceptions import ClientError
from graphlib import TopologicalSorter

from qldebugger.aws import get_account_id, get_client
from qldebugger.config import get_config
from qldebugger.config.file_parser import ConfigQueue, ConfigSecretString

if TYPE_CHECKING:
    from mypy_boto3_sqs.literals import QueueAttributeNameType

logger = logging.getLogger(__name__)


def create_secrets() -> None:
    secretsmanager = get_client('secretsmanager')
    secrets = get_config().secrets

    for name, value in secrets.items():
        try:
            secretsmanager.describe_secret(SecretId=name)
            if isinstance(value, ConfigSecretString):
                logger.info('Updating %r string secret...', name)
                secretsmanager.put_secret_value(SecretId=name, SecretString=value.get_value())
            else:
                logger.info('Updating %r binary secret...', name)
                secretsmanager.put_secret_value(SecretId=name, SecretBinary=value.get_value())
        except ClientError:
            if isinstance(value, ConfigSecretString):
                logger.info('Creating %r string secret...', name)
                secretsmanager.create_secret(Name=name, SecretString=value.get_value())
            else:
                logger.info('Creating %r binary secret...', name)
                secretsmanager.create_secret(Name=name, SecretBinary=value.get_value())


def create_topics() -> None:
    sns = get_client('sns')
    topics = get_config().topics

    for topic_name in topics:
        logger.info('Creating %r topic...', topic_name)
        sns.create_topic(Name=topic_name)


def create_queues() -> None:
    sqs = get_client('sqs')
    queues = get_config().queues
    order = TopologicalSorter(
        {
            name: {queue.redrive_policy.dead_letter_queue} if queue.redrive_policy else set()
            for name, queue in queues.items()
        }
    ).static_order()

    for queue_name in order:
        attributes: Dict['QueueAttributeNameType', str] = {}
        if redrive_policy := queues.get(queue_name, ConfigQueue()).redrive_policy:
            logger.debug('Checking dead letter queue (%r) for %r...', redrive_policy.dead_letter_queue, queue_name)
            dead_letter_queue_attributes = sqs.get_queue_attributes(
                QueueUrl=redrive_policy.dead_letter_queue, AttributeNames=['QueueArn']
            )
            attributes['RedrivePolicy'] = json.dumps(
                {
                    'deadLetterTargetArn': dead_letter_queue_attributes['Attributes']['QueueArn'],
                    'maxReceiveCount': redrive_policy.max_receive_count,
                }
            )
        try:
            queue_url = sqs.get_queue_url(QueueName=queue_name)
            logger.info('Updating %r queue...', queue_name)
            sqs.set_queue_attributes(QueueUrl=queue_url['QueueUrl'], Attributes=attributes)
        except ClientError:
            logger.info('Creating %r queue...', queue_name)
            sqs.create_queue(QueueName=queue_name, Attributes=attributes)


def subscribe_topics() -> None:
    sns = get_client('sns')
    topics = get_config().topics
    account_id = get_account_id()
    region = sns.meta.region_name

    subscriptions = sns.list_subscriptions()['Subscriptions']
    for subscription in subscriptions:
        logger.info('Unsubscribing %r queue (%r)...', subscription['Endpoint'], subscription['SubscriptionArn'])
        sns.unsubscribe(SubscriptionArn=subscription['SubscriptionArn'])

    for topic_name, config_topic in topics.items():
        for subscriber in config_topic.subscribers:
            logger.info('Subscribing %r topic to %r queue...', topic_name, subscriber.queue)
            attributes = {
                'RawMessageDelivery': 'true' if subscriber.raw_message_delivery else 'false',
            }
            if subscriber.filter_policy is not None:
                attributes['FilterPolicy'] = subscriber.filter_policy
            sns.subscribe(
                TopicArn=f'arn:aws:sns:{region}:{account_id}:{topic_name}',
                Protocol='sqs',
                Endpoint=f'arn:aws:sqs:{region}:{account_id}:{subscriber.queue}',
                Attributes=attributes,
            )



import logging
from pathlib import Path
from typing import Optional, Union

from .file_parser import Config

logger = logging.getLogger(__name__)

_current_config: Optional[Config] = None


def load_config(filename: Union[str, Path], /) -> Config:
    logger.debug('Loading %r config...', str(filename))
    global _current_config  # noqa: PLW0603
    with Path(filename).open('rb') as fp:
        _current_config = Config.from_toml(fp)
    return _current_config


def get_config() -> Config:
    if _current_config is None:
        raise RuntimeError('Configuration file is not loaded yet')
    return _current_config



import logging
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Literal, Optional, Union, overload

import boto3
from botocore.config import Config

from .config import get_config

if TYPE_CHECKING:
    from mypy_boto3_secretsmanager import SecretsManagerClient
    from mypy_boto3_sns import SNSClient
    from mypy_boto3_sqs import SQSClient
    from mypy_boto3_sts import STSClient

logger = logging.getLogger(__name__)


@overload
def get_client(service_name: Literal['secretsmanager'], /) -> 'SecretsManagerClient': ...


@overload
def get_client(service_name: Literal['sns'], /) -> 'SNSClient': ...


@overload
def get_client(service_name: Literal['sqs'], /) -> 'SQSClient': ...


@overload
def get_client(service_name: Literal['sts'], /) -> 'STSClient': ...


def get_client(service_name, /):  # type: ignore[no-untyped-def]
    aws_config = get_config().aws
    logger.debug('Connecting to %s service...', service_name)
    session = boto3.Session(
        aws_access_key_id=aws_config.access_key_id,
        aws_secret_access_key=aws_config.secret_access_key,
        aws_session_token=aws_config.session_token,
        region_name=aws_config.region,
        profile_name=aws_config.profile,
    )
    return session.client(
        service_name=service_name,
        endpoint_url=aws_config.endpoint_url,
    )


get_client = lru_cache(maxsize=None)(get_client)  # type: ignore[assignment]


def get_account_id() -> str:
    sts = get_client('sts')
    return sts.get_caller_identity()['Account']


def inject_aws_config_in_client(
    service_name: str,
    region_name: Optional[str] = None,
    api_version: Optional[str] = None,
    use_ssl: Optional[bool] = True,
    verify: Optional[Union[bool, str]] = None,
    endpoint_url: Optional[str] = None,
    aws_access_key_id: Optional[str] = None,
    aws_secret_access_key: Optional[str] = None,
    aws_session_token: Optional[str] = None,
    config: Optional[Config] = None,
) -> Any:
    aws_config = get_config().aws
    if aws_config.access_key_id is not None:
        aws_access_key_id = aws_config.access_key_id
    if aws_config.secret_access_key is not None:
        aws_secret_access_key = aws_config.secret_access_key
    if aws_config.session_token is not None:
        aws_session_token = aws_config.session_token
    if aws_config.region is not None:
        region_name = aws_config.region
    if aws_config.endpoint_url is not None:
        endpoint_url = aws_config.endpoint_url
    session = boto3.Session(
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key,
        aws_session_token=aws_session_token,
        region_name=region_name,
        profile_name=aws_config.profile,
    )
    return session.client(  # type: ignore[call-overload,misc]
        service_name=service_name,
        api_version=api_version,
        use_ssl=use_ssl,
        verify=verify,
        endpoint_url=endpoint_url,
        config=config,
    )



import logging
from importlib import import_module
from traceback import format_exc
from typing import TYPE_CHECKING, Any, Callable
from unittest.mock import patch

from qldebugger.aws import inject_aws_config_in_client
from qldebugger.config import get_config

if TYPE_CHECKING:
    from aws_lambda_typing.events import SQSEvent

logger = logging.getLogger(__name__)


def get_lambda_function(*, lambda_name: str) -> Callable[['SQSEvent', None], Any]:
    module_name, function_name = get_config().lambdas[lambda_name].handler
    logger.debug('Importing lambda_handler of %r...', lambda_name)
    lambda_handler: Callable[['SQSEvent', None], Any] = getattr(import_module(module_name), function_name)
    return lambda_handler


def run_lambda(*, lambda_name: str, event: 'SQSEvent') -> Any:
    environment = get_config().lambdas[lambda_name].environment
    with patch('boto3.client', inject_aws_config_in_client):
        lambda_handler = get_lambda_function(lambda_name=lambda_name)
        logger.info('Running %r lambda...', lambda_name)
        try:
            with patch.dict('os.environ', environment, True):
                result = lambda_handler(event, None)
        except Exception:
            logger.exception('Error on execute lambda:\n%s', format_exc())
            raise
        logger.info('Result: %r', result)
        return result



import json
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Mapping

import click

from . import actions
from .config import load_config

if TYPE_CHECKING:
    from mypy_boto3_sns.type_defs import MessageAttributeValueTypeDef

CONFIG_FILENAME = Path('qldebugger.toml')


@click.group()
def cli() -> None:
    logging.basicConfig(level=logging.INFO, format='%(levelname)s:qldebugger:%(message)s')


@cli.command()
def init() -> None:
    config = """[aws]
profile = ""
access_key_id = "secret"
secret_access_key = "secret"
session_token = ""
region = "us-east-1"
endpoint_url = "http://localhost:4566/"

[queues]
myqueue = {}

[lambdas]
print = {handler = "qldebugger.example.lambdas.print_messages"}
[lambdas.fail]
handler = "qldebugger.example.lambdas.exec_fail"
[lambdas.fail.environment]
VARIABLE = "VALUE"

[event_source_mapping]
a = {queue = "myqueue", function_name = "print"}
b = {queue = "myqueue", function_name = "fail", batch_size = 1, maximum_batching_window = 20}
"""
    if CONFIG_FILENAME.exists():
        click.echo('Configuration file already exists')
        return
    with CONFIG_FILENAME.open('w') as fp:
        fp.write(config)
    click.echo('Configuration file created')


@cli.command()
@click.argument('event_source_mapping_name')
def run(event_source_mapping_name: str) -> None:
    load_config(CONFIG_FILENAME)
    actions.event_source_mapping.receive_messages_and_run_lambda(event_source_mapping_name=event_source_mapping_name)


# Infra


@cli.group()
def infra() -> None: ...


@infra.command('create-secrets')
def infra_create_secrets() -> None:
    load_config(CONFIG_FILENAME)
    actions.infra.create_secrets()


@infra.command('create-topics')
def infra_create_topics() -> None:
    load_config(CONFIG_FILENAME)
    actions.infra.create_topics()


@infra.command('create-queues')
def infra_create_queues() -> None:
    load_config(CONFIG_FILENAME)
    actions.infra.create_queues()


@infra.command('create-all')
def infra_create_all() -> None:
    load_config(CONFIG_FILENAME)
    actions.infra.create_secrets()
    actions.infra.create_topics()
    actions.infra.create_queues()
    actions.infra.subscribe_topics()


@infra.command('subscribe-topics')
def infra_subscribe_topics() -> None:
    load_config(CONFIG_FILENAME)
    actions.infra.subscribe_topics()


# Msg


@cli.group()
def msg() -> None: ...


@msg.command('publish')
@click.argument('topic_name')
@click.argument('message')
@click.argument('attributes', default='{}', type=json.loads)
def msg_publish(topic_name: str, message: str, attributes: Mapping[str, 'MessageAttributeValueTypeDef']) -> None:
    load_config(CONFIG_FILENAME)
    actions.message.publish_message(topic_name=topic_name, message=message, attributes=attributes)


@msg.command('send')
@click.argument('queue_name')
@click.argument('message')
def msg_send(queue_name: str, message: str) -> None:
    load_config(CONFIG_FILENAME)
    actions.message.send_message(queue_name=queue_name, message=message)


@msg.command('receive')
@click.argument('queue_name')
@click.option('--batch-size', default=1, type=int, show_default=True)
@click.option('--wait-seconds', default=0, type=int, show_default=True)
def msg_receive(queue_name: str, batch_size: int, wait_seconds: int) -> None:
    load_config(CONFIG_FILENAME)
    messages = actions.message.receive_message(
        queue_name=queue_name, batch_size=batch_size, maximum_batching_window=wait_seconds
    )
    for message in messages['Messages']:
        click.echo(repr(message.get('Body')))
    actions.message.delete_messages(queue_name=queue_name, messages=messages)


@msg.command('purge')
@click.argument('queue_name')
def msg_purge(queue_name: str) -> None:
    load_config(CONFIG_FILENAME)
    actions.message.purge_messages(queue_name=queue_name)
